/*
 * @(#)UnotOD.java
 *
 * Copyright 2003, Tatsuya Asai, Hiroki Arimura, Takeaki Uno,
 * and Shin-ichi Nakano, all rights reserved.
 *
 * This file contains an implementation of the algorithm "UNOT".
 *
 * @see Discovering Frequent Substructures in Large Unordered Trees,
 * In Proc. the 6th Discovery Science (DS'03), LNAI, October 2003.
 * <http://www.i.kyushu-u.ac.jp/doitr/trcs216.pdf>
 *
 * This software may be used freely for any purpose. However, when distributed,
 * the original source must be clearly stated, and, when the source code is
 * distributed, the copiright notice must be retained and any alterations in
 * the code must be clearly marked. No warranty is given regarding the quality
 * of this software.
 */
package unot;

import java.util.*;

/**
 * UnotOD.java
 *
 * `OD' means `Occurrence Deliver'.
 *
 * Created: Sat Feb  5 03:04:00 2005
 *
 * @author <a href="mailto:s-fusa@i.kyushu-u.ac.jp">Shinji Fusanobu</a>
 * @version $$
 * @see <a href="http://www.i.kyushu-u.ac.jp/doitr/trcs216.pdf">
 * Discovering Frequent Substructures in Large Unordered Trees (PDF)</a>
 */
public class UnotNonOD extends Unot
{

	static final boolean ENABLE_DELETE_NONCANONICAL_OCC = false; 

    /**
     * Creates a new <code>UnotNonOD</code> instance.
     *
     */
    protected UnotNonOD()
    {
        super();
    }


    /**
     * Creates a new <code>UnotNonOD</code> instance.
     *
     * @param docset a <code>DocSet</code> value
     */
    public UnotNonOD(DocSet docset)
    {
        super(docset);
    } // Unot constructor


    /**
     * Describe <code>expand</code> method here.
     *
     * @param pattern an <code>UnotPattern</code> value
     * @param occList an <code>OccList</code> value
     * @param minSup an <code>int</code> value
     */
    //protected void expand(UnotPattern pattern, OccList occList, int minSup)


    /**
     * Expands the <code>UnotPattern</code> and calls itself
     * if the expanded patterns is embeded in Documents more than the thresholds.
     *
     * @param pattern an <code>UnotPattern</code> value
     * @param occList an <code>OccList</code> value
     * @param minSup an <code>int</code> value
     */
    protected void expand(UnotPattern pattern, OccList occList, int minSup)
    {

        this.numberOfExpansion++;

        // 祳ɥۤƤкƵ
        if ((this.maxCodeSize != MAX_SIZE_NO_LIMIT)
            && (this.maxCodeSize < pattern.getCodeSize()))
            return;

        // ѽФǤСΥѥä
        if (this.docset.getOccSize(occList) < minSup) {
            return;
        }

		//ȤʤѽХѥ󤬸Ĥä
		maintainPatternInfo(pattern, occList);

		this.patset.add(pattern);

        // ʥѥ
        List newPatterns = findAllChildren(pattern);

        for (int i = 0; i < newPatterns.size(); i++) {
            OccList newOccList = updateOcc((UnotPattern) newPatterns.get(i), occList);
            if (newOccList.size() > 0) {
                this.numberOfOccurrence += newOccList.size();
                expand((UnotPattern) newPatterns.get(i), newOccList, minSup);
            }
        }

    } // expand


    /**
     * Finds all child patterns expanded with Rightmost Expansion
     * from one <code>UnotPattern</code>.
     *
     * @param pattern an <code>UnotPattern</code> value
     * @return a <code>List</code> value
     */
    protected List findAllChildren(UnotPattern pattern)
    {

        int copyDepth = pattern.getCopyDepth();
        List unotPatternList = new ArrayList();

        if (pattern.isBalanced(copyDepth)) { // if Lc == Rc
            for (int dep = 1; dep <= copyDepth+1; dep++) {
                int maxLabel = pattern.getNodeLabel(dep);

                for (int lab = 0; lab < maxLabel; lab++) {
                    UnotPattern newPattern =
                        pattern.rightmostExpand(pattern, new DLpair(dep, lab), dep);
                    unotPatternList.add(newPattern);
                }

                if (! pattern.isInftyNode(dep)) {
                    UnotPattern newPattern =
                        pattern.rightmostExpand(pattern, new DLpair(dep, maxLabel), dep-1);
                    unotPatternList.add(newPattern);
                }
            }

        } else { // if Lc != Rc
            /* `da node' means `disagreement node'. */
            int daNodeDepth = pattern.getDisagreeNodeDepth(copyDepth);
            int daNodeLabel = pattern.getDisagreeNodeLabel(copyDepth);

            // when ( 1 <= dep < daNodeDepth )
            for (int dep = 1; dep < daNodeDepth; dep++) {
                int maxLabel = pattern.getNodeLabel(dep);

                for (int lab = 0; lab < maxLabel; lab++) {
                    UnotPattern newPattern =
                        pattern.rightmostExpand(pattern, new DLpair(dep, lab), dep);
                    unotPatternList.add(newPattern);
                }

                UnotPattern newPattern =
                    pattern.rightmostExpand(pattern, new DLpair(dep, maxLabel), dep-1);
                unotPatternList.add(newPattern);
            }

            // when ( dep == daNodeDepth )
            for (int lab = 0; lab < daNodeLabel; lab++) {
                UnotPattern newPattern =
                    pattern.rightmostExpand(pattern, new DLpair(daNodeDepth, lab), daNodeDepth);
                unotPatternList.add(newPattern);
            }

            UnotPattern newPattern =
                pattern.rightmostExpand(pattern, new DLpair(daNodeDepth, daNodeLabel), copyDepth);
            unotPatternList.add(newPattern);
        }

        return unotPatternList;

    } // findAllChildren


    /**
     * Updates the old Occurrence List to the current Occurrence List
     * with the expanded <code>UnotPattern</code>.
     *
     * @param newPattern an <code>UnotPattern</code> value
     * @param oldOccList a <code>List</code> value
     * @return a <code>List</code> value
     */
    protected OccList updateOcc(UnotPattern newPattern, OccList oldOccList)
    {

        OccList newOccList = new OccList();

        int copyDepth = newPattern.getCopyDepth();
        int topDepth = newPattern.getTopDepth();
        int topLabel = newPattern.getNodeLabel(topDepth);
        int parentIndex = newPattern.getNodeIndex(topDepth -1);

        for (int i = 0; i < oldOccList.size(); i++) {

            Occurrence occ = (Occurrence) oldOccList.get(i);
            Occurrence newOcc = null;

            UnotNode parent = (UnotNode) occ.get(parentIndex);
            NodeList children = parent.getChildren();

            for (int j = 0; j < children.size(); j++) {
                UnotNode child = (UnotNode) children.get(j);
                if ((child.getLabelId() == topLabel) && ! occ.contains(child)) {
                    newOcc = new Occurrence(occ, child);
					if (!useBalanceCheck || //if useBalanceCheck is false then add
						isCanonicalOcc(newPattern, newOcc, copyDepth, topDepth)) {
						//Add new occurrence
                        newOccList.add(newOcc);
                    }
                }
            }
        }

        return newOccList;

    } // updateOcc

    /**
     * ̵ڤ󥢥르ꥺ findAllChildren() ñȤǺƵŪ˼¹Ԥ롥
     *
     * @param pattern an <code>UnotPattern</code> value
     * @param maxLabel an <code>int</code> value
     * @param maxSize an <code>int</code> value
     */
    public void enumerate(UnotPattern pattern, int maxLabel, int maxSize)
    {

        if (maxSize < pattern.getCodeSize())
            return;

        this.patset.add(pattern);

        UnotLog.log("@pattern: " + this.patset.size() + "\n");
        UnotLog.log(pattern + "\n");
        UnotLog.log(pattern.writeToString() + "\n\n");
        List newPatterns = findAllChildren(pattern);
        for (int i = 0; i < newPatterns.size(); i++) {
            enumerate((UnotPattern) newPatterns.get(i), maxLabel, maxSize);
        }
    }


    /**
     * findAllChildren() ΥƥȥɡѥΤߤԤ
     *
     * @param args a <code>String[]</code> value
     */
    public static void main(String[] args)
    {
        if (args.length < 2) {
            UnotLog.log("USAGE: java unot.Unot MAXLABEL MAXSIZE\n");
            return;
        }

        UnotNonOD unot = new UnotNonOD();
        int maxLabel = Integer.parseInt(args[0]);
        int maxSize = Integer.parseInt(args[1]);
        UnotPattern.setMaxLabel(maxLabel);

        for (int i = 0; i < maxLabel; i++) {
            unot.enumerate(new UnotPattern(i), maxLabel, maxSize);
        }

        UnotLog.log("@total-pattern: " + unot.patset.size() + "\n");
    }

} // UnotNonOD
